@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012-2013 Fanout, Inc.
+ * Copyright (C) 2012-2016 Fanout, Inc.
  * 
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -33,9 +33,13 @@ void log_setOutputLevel(int level);
 bool log_setFile(const QString &fname);
 bool log_rotate();
 
+void log(int level, const char *fmt, ...);
 void log_error(const char *fmt, ...);
 void log_warning(const char *fmt, ...);
 void log_info(const char *fmt, ...);
 void log_debug(const char *fmt, ...);
 
+// log without prefixing or anything. useful for forwarding log data
+void log_raw(const char *line);
+
 #endif
