@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012-2013 Fanout, Inc.
+ * Copyright (C) 2012-2016 Fanout, Inc.
  * 
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -27,8 +27,31 @@ QVariant ZhttpRequestPacket::toVariant() const
 	if(!from.isEmpty())
 		obj["from"] = from;
 
-	if(!id.isEmpty())
-		obj["id"] = id;
+	if(!ids.isEmpty())
+	{
+		if(ids.count() == 1)
+		{
+			const Id &id = ids.first();
+			if(!id.id.isEmpty())
+				obj["id"] = id.id;
+			if(id.seq != -1)
+				obj["seq"] = id.seq;
+		}
+		else
+		{
+			QVariantList vl;
+			foreach(const Id &id, ids)
+			{
+				QVariantHash vh;
+				if(!id.id.isEmpty())
+					vh["id"] = id.id;
+				if(id.seq != -1)
+					vh["seq"] = id.seq;
+				vl += vh;
+			}
+			obj["id"] = vl;
+		}
+	}
 
 	QByteArray typeStr;
 	switch(type)
@@ -51,9 +74,6 @@ QVariant ZhttpRequestPacket::toVariant() const
 	if(type == Error && !condition.isEmpty())
 		obj["condition"] = condition;
 
-	if(seq != -1)
-		obj["seq"] = seq;
-
 	if(credits != -1)
 		obj["credits"] = credits;
 
@@ -125,6 +145,22 @@ QVariant ZhttpRequestPacket::toVariant() const
 	if(followRedirects)
 		obj["follow-redirects"] = true;
 
+	if(passthrough.isValid())
+		obj["passthrough"] = passthrough;
+
+	if(multi || quiet)
+	{
+		QVariantHash ext;
+
+		if(multi)
+			ext["multi"] = true;
+
+		if(quiet)
+			ext["quiet"] = true;
+
+		obj["ext"] = ext;
+	}
+
 	return obj;
 }
 
@@ -144,12 +180,59 @@ bool ZhttpRequestPacket::fromVariant(const QVariant &in)
 		from = obj["from"].toByteArray();
 	}
 
+	ids.clear();
 	if(obj.contains("id"))
 	{
-		if(obj["id"].type() != QVariant::ByteArray)
+		if(obj["id"].type() == QVariant::ByteArray)
+		{
+			Id id;
+			id.id = obj["id"].toByteArray();
+			ids += id;
+		}
+		else if(obj["id"].type() == QVariant::List)
+		{
+			QVariantList vl = obj["id"].toList();
+			foreach(const QVariant &v, vl)
+			{
+				if(v.type() != QVariant::Hash)
+					return false;
+
+				Id id;
+
+				QVariantHash vh = v.toHash();
+
+				if(vh.contains("id"))
+				{
+					if(vh["id"].type() != QVariant::ByteArray)
+						return false;
+
+					id.id = vh["id"].toByteArray();
+				}
+
+				if(vh.contains("seq"))
+				{
+					if(!vh["seq"].canConvert(QVariant::Int))
+						return false;
+
+					id.seq = vh["seq"].toInt();
+				}
+
+				ids += id;
+			}
+		}
+		else
+			return false;
+	}
+
+	if(obj.contains("seq"))
+	{
+		if(!obj["seq"].canConvert(QVariant::Int))
 			return false;
 
-		id = obj["id"].toByteArray();
+		if(ids.isEmpty())
+			ids += Id();
+
+		ids.first().seq = obj["seq"].toInt();
 	}
 
 	type = Data;
@@ -194,15 +277,6 @@ bool ZhttpRequestPacket::fromVariant(const QVariant &in)
 		}
 	}
 
-	seq = -1;
-	if(obj.contains("seq"))
-	{
-		if(!obj["seq"].canConvert(QVariant::Int))
-			return false;
-
-		seq = obj["seq"].toInt();
-	}
-
 	credits = -1;
 	if(obj.contains("credits"))
 	{
@@ -312,7 +386,7 @@ bool ZhttpRequestPacket::fromVariant(const QVariant &in)
 		code = obj["code"].toInt();
 	}
 
-	userData = obj["user-data"];
+	userData = obj.value("user-data");
 
 	peerAddress = QHostAddress();
 	if(obj.contains("peer-address"))
@@ -386,5 +460,25 @@ bool ZhttpRequestPacket::fromVariant(const QVariant &in)
 		followRedirects = obj["follow-redirects"].toBool();
 	}
 
+	passthrough = obj.value("passthrough");
+
+	multi = false;
+	if(obj.contains("ext"))
+	{
+		if(obj["ext"].type() != QVariant::Hash)
+			return false;
+
+		QVariantHash ext = obj["ext"].toHash();
+		if(ext.contains("multi") && ext["multi"].type() == QVariant::Bool)
+		{
+			multi = ext["multi"].toBool();
+		}
+
+		if(ext.contains("quiet") && ext["quiet"].type() == QVariant::Bool)
+		{
+			quiet = ext["quiet"].toBool();
+		}
+	}
+
 	return true;
 }
