@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012-2013 Fanout, Inc.
+ * Copyright (C) 2012-2016 Fanout, Inc.
  * 
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -26,6 +26,24 @@
 class ZhttpRequestPacket
 {
 public:
+	class Id
+	{
+	public:
+		QByteArray id;
+		int seq;
+
+		Id() :
+			seq(-1)
+		{
+		}
+
+		Id(const QByteArray &_id, int _seq = -1) :
+			id(_id),
+			seq(_seq)
+		{
+		}
+	};
+
 	enum Type
 	{
 		Data,
@@ -41,11 +59,10 @@ public:
 	};
 
 	QByteArray from;
-	QByteArray id;
+	QList<Id> ids;
 
 	Type type;
 	QByteArray condition;
-	int seq;
 
 	int credits;
 	bool more;
@@ -72,10 +89,12 @@ public:
 	bool trustConnectHost;
 	bool ignoreTlsErrors;
 	bool followRedirects;
+	QVariant passthrough; // if valid, may contain pushpin-specific passthrough info
+	bool multi;
+	bool quiet;
 
 	ZhttpRequestPacket() :
 		type((Type)-1),
-		seq(-1),
 		credits(-1),
 		more(false),
 		stream(false),
@@ -87,7 +106,9 @@ public:
 		ignorePolicies(false),
 		trustConnectHost(false),
 		ignoreTlsErrors(false),
-		followRedirects(false)
+		followRedirects(false),
+		multi(false),
+		quiet(false)
 	{
 	}
 
