@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012-2013 Fanout, Inc.
+ * Copyright (C) 2012-2017 Fanout, Inc.
  * 
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,7 +21,9 @@
 #include <QObject>
 
 class QVariant;
-class JDnsShared;
+class QJDnsShared;
+class ZhttpRequestPacket;
+class ZhttpResponsePacket;
 class AppConfig;
 
 class Worker : public QObject
@@ -41,17 +43,17 @@ public:
 		JsonFormat
 	};
 
-	Worker(JDnsShared *dns, AppConfig *config, Format format, QObject *parent = 0);
+	Worker(QJDnsShared *dns, AppConfig *config, Format format, QObject *parent = 0);
 	~Worker();
 
 	QByteArray rid() const;
 	Format format() const;
 
-	void start(const QVariant &request, Mode mode);
-	void write(const QVariant &request);
+	void start(const QByteArray &id, int seq, const ZhttpRequestPacket &request, Mode mode);
+	void write(int seq, const ZhttpRequestPacket &request);
 
 signals:
-	void readyRead(const QByteArray &receiver, const QVariant &response);
+	void readyRead(const QByteArray &receiver, const ZhttpResponsePacket &response);
 	void finished();
 
 private:
